/* *** Preamble *** */

/*jslint browser: true, devel: true */
/*global d3 */
/*global numeric */


/*  *** Introduction ***

    The relationship between points in the plane and gambles is as follows:

    (i) A gamble that incurs sure loss corresponds to a point inside a regular
        hexagon (with ‘radius’ r0) symetrically positioned around the origin.
    (ii) A gamble that incurs sure gain corresponds to a point outside a
         strictly larger such hexagon (with ‘radius’ r1).
    (iii) The gambles of interest lie between the two hexagons.

    The ‘radius’ of such a hexagon is the distance of the origin to one of the
    hexagon's corners. Each of the corners, or, rather, each of the axes
    defined by the hexagons' corners corresponds to an event. We have the
    following (event, angle)-pairs:

        Win, -5π/6
        Not Draw, -π/2
        Lose, -π/6
        Not Win, π/6
        Draw, π/2
        Not Lose, 5π/6

    The distance between the two hexagons is normalized to a mixture parameter
    β between 0 (corresponding to r0, which could be interpreted as full
    confidence) and 1 (corresponding to r1, which could be interpreted as no
    confidence). The angle is related to gambles as follows: given a state
    (Win, Draw, or Lose) and a negated state (No Win, No Draw, or No Lose), the
    angle within the quadrangle delimited by the corresponding axes is
    translated into a second mixture parameter, λ between 0 (corresponding to
    the state) and 1 (corresponding to the negated state).

    Now, these mixture parameters do not directly translate to gamble values,
    but actually to binary logarithms (lg) of gamble values. The reason is that
    we wish to be able to express a wide range of relative values in the
    gambles rather than high precision is a limited range, as the potential
    winnings or losses relative to the stake (highest absolute loss, normalized
    to 1 when translating to planar coordinates) are important.

    The range of absolute values considered (apart from 0) is
    [2^-lg.inf, 2^lg.inf], where lg.inf is a chosen parameter that replaces actual
    Infinity, i.e., everything close to lg.inf (beyond lgthresh) is considered
    to *be* lg.inf. The full range of (positive and negative) numbers considered
    is

        [-1, ..., -2^-lg.inf, 0, 2^-lg.inf, ..., 2^lg.inf].

    Any gamble between the hexagons has a negative minimal value (-stake, so -1
    when normalized), a nonnegative maximal value (called the odds), and a mid
    value in between these (or equal to one of these). The relationship between
    β, λ, the state, and the negation state and these values is as follows: the
    state whose negation is included corresponds to -stake, the included state
    corresponds to the odds, and the remaining state's value is determined by
    λ. We have the relationships

        stake = 1,
        β = (lg(odds) + lg.inf) / (2 * lg.inf),
        λ = σ / (lg(odds) + 2 * lg.inf),

    where σ is an auxiliary parameter:

        σ = -lg(-mid)  if mid < 0.
        σ = lg.inf  if mid = 0.
        σ = 2 lg.inf + lg(mid)  if mid > 0.

    This approach guarantees that binary logarithms of absolute gamble values
    vary linearly in the visualizaton. We use a discrete number of gambles in
    the visualizaton, and so these gambles are spaced in a uniform grid even
    though their values differ by an order of 2 (or infinity, when the value
    goes to 0).

    In the visualizaton, gambles with infinite odds are not considered for
    selection by the user, as these are a proxy for nowhere negative gambles,
    which are implicitly acceptable for everyone, implying no possibly of loss.

                                                                            */


var GambleHex = {};  // The namespace for global objects

(function () { // Default names for events
    "use strict";
    var not = "Not",
        lbl;
    GambleHex.lbl = {};
    lbl = GambleHex.lbl;
    lbl.W = "Win";
    lbl.D = "Draw";
    lbl.L = "Lose";
    lbl.NW = not + " " + lbl.W;
    lbl.ND = not + " " + lbl.D;
    lbl.NL = not + " " + lbl.L;
}());

GambleHex.sgn = function (val) {  // signum function
    "use strict";
    if (isNaN(val)) { throw val + 'is NaN'; }
    return val === 0 ? val : (val < 0 ? -1 : 1);
};


/* Define a saturating lg (binary logarithm) function
   Range of values considered is -1 ... -2^-lg.max 0 2^-lg.max ... 2^lg.max */
GambleHex.lg = {};
(function () {
    "use strict";
    var lg = GambleHex.lg;
    lg.max = 3; // should be at least 1
    lg.inf = lg.max + 1;
    lg.saturate = function (val) {
        var thresh = lg.max + 0.5,
            lgval = Math.log(val) / Math.LN2;
        if (Math.abs(lgval) > thresh) { return lgval > 0 ? lg.inf : -lg.inf; }
        return lgval;
    };
}());
GambleHex.twoexp = function (val) {  // powers of two
    "use strict";
    return Math.pow(2, val);
};


/* Relative distances and location parameters of the visualizaton */
GambleHex.offset = 200;
        // this determines the sizing of the visualizaton, together with the
        // svg width, height, and viewbox attributes (see below)
GambleHex.origin = {'x': GambleHex.offset, 'y': GambleHex.offset};


/* *** Definition of a gamble and the coordinates that correspond to it *** */

/* Gamble as d3.map from {'W', 'D', 'L'} to Number */
GambleHex.normalize = function (w, d, l) { // normalize so that stake = 1
    "use strict";
    var sgn = GambleHex.sgn,
        gamble = d3.map({'W': w, 'D': d, 'L': l}),
        vals = gamble.values().sort(d3.ascending),
        stake = -vals[0],
        odds = vals[2],
        mid = vals[1];
    if (odds < 0) { throw "sure loss"; }
    switch (sgn(stake)) {
    case -1:
        throw "sure gain";
    case 0:
        if (odds === 0) { throw "status quo"; }
        gamble.forEach(function (key, val) {
            gamble.set(key, val > 0 ? +Infinity : -1);
        });
        break;
    case +1:
        if (odds === 0) {
            if ((mid !== -stake) || (stake !== 1)) {  // not yet normalized
                gamble.forEach(function (key, val) {
                    gamble.set(key, val < 0 ? -1 : 0);
                });
            }
        } else {
            if (stake !== 1) {  // not yet normalized
                gamble.forEach(function (key, val) {
                    gamble.set(key, val / stake);
                });
            }
        }
        break;
    }
    return gamble;
};

/* SVG coordinates corresponding to a gamble */
GambleHex.coords = function (w, d, l) {
    "use strict";
    var sgn = GambleHex.sgn,
        lg = GambleHex.lg,
        gamble = GambleHex.normalize(w, d, l),
        vals = gamble.values().sort(d3.ascending),
        odds = vals[2],
        lgodds = lg.saturate(odds),
        mid = vals[1],
        beta,  // mixture coefficient between odds 0 and +Infinity
        sigma,
        lambda,  //  mixture coefficient between mid state and odds state
        r,  // hexagon-radial distance from origin
        r1 = 3 * GambleHex.offset / 5, // β = 1
        r0 = r1 / 3,  // β = 0 (always three times as many ‘dots’ at lg.inf)
        rnorm = r1 - r0,
        rho,  // radial distance from origin
        psi,  // angle with closest odds state ('W', 'D', or 'L')
        theta;  // polar coordinate angle
    /* Calculate the gamble's mixture coefficients */
    beta = (lgodds + lg.inf) / (2 * lg.inf);
        // odds = 0 => β = 0, odds = +Infinity => β = 1
    switch (sgn(mid)) {
    case -1:
        sigma = -lg.saturate(-mid);
        break;
    case 0:
        sigma = lg.inf;
        break;
    case +1:
        sigma = lg.saturate(mid) + 2 * lg.inf;
        break;
    }
    lambda = sigma / (lgodds + 2 * lg.inf);
        // mid -> -stake => λ -> 0, mid -> odds => λ -> 1
    /* Calculate polar coordinates ρ, θ */
    r = r0 + beta * rnorm;
    /* ρ and ψ are calculated by solving the triangle

        ((ρ, π/3), (λr, ψ), (r, 2π/3-ψ))

    given in (side length, opposite angle)-notation                         */
    rho = r * Math.sqrt(1 - lambda * (1 - lambda));
    psi = Math.asin(lambda * (Math.sqrt(3) / 2) * (r / rho));
    /* θ is determined by the odds state and the stake state */
    function sign_psi(state) { return gamble.get(state) === -1 ? psi : -psi; }
    switch (odds) {
    case gamble.get('W'):
        theta = -5 * Math.PI / 6 + sign_psi('D');
        break;
    case gamble.get('D'):
        theta = Math.PI / 2 + sign_psi('L');
        break;
    case gamble.get('L'):
        theta = -Math.PI / 6 + sign_psi('W');
        break;
    }
    /* Return the coordinates in the SVG coordinate system */
    return {'beta': beta,
            'lambda': lambda,
            'x': GambleHex.origin.x + rho * Math.cos(theta),
            'y': GambleHex.origin.y - rho * Math.sin(theta)};
                // SVG y axis points downwards
};


/* Gambles that can be accepted */
GambleHex.grid = d3.map();
GambleHex.griddata = d3.map();
GambleHex.dec2frac = function (val) {  // Transform gamble values into strings
    "use strict";
    var sgn = GambleHex.sgn,
        lg = GambleHex.lg,
        twoexp = GambleHex.twoexp,
        sign = sgn(val);
    if (val === 0) { return "±0"; }
    sign = sgn(val) === +1 ? '＋' : '−';
        // we use unicode full width plus and minus in id names to avoid
        // invalid character problems is css selectors
    val = Math.abs(val);
    if (val >= 1) { return sign + twoexp(lg.saturate(val)); }
    switch (val) {
    case 0.5:
        return sign + '½';
    case 0.25:
        return sign + '¼';
    case 0.125:
        return sign + '⅛';
    default:
        return sign + '1÷' + (1 / twoexp(lg.saturate(val)));
                        // unicode obelus to avoid ...
    }
};
(function () {  // Generate the grid
    "use strict";
    var sgn = GambleHex.sgn,
        lg = GambleHex.lg,
        twoexp = GambleHex.twoexp,
        normalize = GambleHex.normalize,
        grid = GambleHex.grid,
        griddata = GambleHex.griddata,
        dec2frac = GambleHex.dec2frac,
        valrange = [].concat(
            d3.range(-lg.max, 1).map(function (lgval) {
                return -twoexp(lgval);
            }),
            [0],
            d3.range(-lg.max, +lg.inf).map(function (lgval) {
                return twoexp(lgval);
            })
        );
    /* Generate the grid of gambles that can be accepted */
    valrange.forEach(function (w) {
        valrange.forEach(function (d) {
            valrange.forEach(function (l) {
                var gamble,
                    wval,
                    dval,
                    lval,
                    data,
                    id = '';
                if ((d3.min([w, d, l]) > -1) || (d3.max([w, d, l]) < 0)) {
                    return;  // gamble cannot be normalized
                }
                gamble = normalize(w, d, l);
                wval = gamble.get('W');
                dval = gamble.get('D');
                lval = gamble.get('L');
                id = id.concat(dec2frac(wval), dec2frac(dval), dec2frac(lval));
                if (grid.has(id)) { return; } // gamble already added to grid
                data = GambleHex.coords(wval, dval, lval);
                griddata.set(id, {
                    'x': data.x,
                    'y': data.y,
                    'gamble': gamble,
                    'dominating': d3.set(),
                    'dominated': d3.set(),
                    'negdominating': d3.set()
                });
                grid.set(id, {
                    'id': id,
                    'beta': data.beta,
                    'lambda': data.lambda,
                    'W': wval,
                    'D': dval,
                    'L': lval
                });
            });
        });
    });
    /* What value present in the grid lies up or down? */
    function val_up(val) {
        var lgval = lg.saturate(Math.abs(val));
        switch (sgn(val)) {
        case -1:
            if (val < -1) { throw "illegal value " + val + " for 'val_up'"; }
            return lgval <= -lg.max ? 0 : val / 2;
        case 0:
            return twoexp(-lg.max);
        case +1:
            if (lgval <= -lg.inf) { return twoexp(-lg.max); }
            return lgval >= lg.max ? undefined : val * 2;
        }
    }
    function val_down(val) {
        var lgval = lg.saturate(Math.abs(val));
        switch (sgn(val)) {
        case -1:
            if (lgval <= -lg.inf) { return -twoexp(-lg.max); }
            return val <= -1 ? undefined : val * 2;
        case 0:
            return -twoexp(-lg.max);
        case +1:
            if (lgval >= lg.inf) { return twoexp(lg.max); }
            return lgval <= -lg.max ? 0 : val / 2;
        }
    }
    /* Calculate neighboring (in direction chosen) gambles in grid
    of a gamble with specified values */
    function neighbors(val_dir, w, d, l) {
        var val,
            wstr = dec2frac(w),
            dstr = dec2frac(d),
            lstr = dec2frac(l),
            ids = [];
        val = val_dir(w);
        if (val !== undefined) {
            ids.push(''.concat(dec2frac(val), dstr, lstr));
        }
        val = val_dir(d);
        if (val !== undefined) {
            ids.push(''.concat(wstr, dec2frac(val), lstr));
        }
        val = val_dir(l);
        if (val !== undefined) {
            ids.push(''.concat(wstr, dstr, dec2frac(val)));
        }
        return ids;
    }
    function negdominated(negw, negd, negl) {
        var gamble;
        gamble = normalize(
            negw === -1 ? -1 : val_down(negw),
            negd === -1 ? -1 : val_down(negd),
            negl === -1 ? -1 : val_down(negl)
        );
        return ''.concat(
            dec2frac(gamble.get('W')),
            dec2frac(gamble.get('D')),
            dec2frac(gamble.get('L'))
        );
    }
    /* Add dominating/dominated/negdominated */
    function ingrid(id) { return grid.has(id); }
    griddata.forEach(function (id, datum) {
        var gamble = datum.gamble,
            w = gamble.get('W'),
            d = gamble.get('D'),
            l = gamble.get('L'),
            neggamble = normalize(-w, -d, -l),
            negw = neggamble.get('W'),
            negd = neggamble.get('D'),
            negl = neggamble.get('L');
        datum.dominating = d3.set(
            neighbors(val_up, w, d, l).filter(ingrid)
        );
        datum.dominated = d3.set(
            neighbors(val_down, w, d, l).filter(ingrid)
        );
        griddata.get(negdominated(negw, negd, negl)).negdominating.add(id);
    });
    /* Add sure dominating gambles of sure gambles */
    griddata.get("±0−1−1").dominating.add("±0±0−1");
    griddata.get("±0−1−1").dominating.add("±0−1±0");
    griddata.get("−1±0−1").dominating.add("±0±0−1");
    griddata.get("−1±0−1").dominating.add("−1±0±0");
    griddata.get("−1−1±0").dominating.add("±0−1±0");
    griddata.get("−1−1±0").dominating.add("−1±0±0");
    /* Add sure dominated gambles of sure gambles */
    griddata.get("±0±0−1").dominated.add("±0−1−1");
    griddata.get("±0−1±0").dominated.add("±0−1−1");
    griddata.get("±0±0−1").dominated.add("−1±0−1");
    griddata.get("−1±0±0").dominated.add("−1±0−1");
    griddata.get("±0−1±0").dominated.add("−1−1±0");
    griddata.get("−1±0±0").dominated.add("−1−1±0");
}());


/*  Deductive inference functions  */
GambleHex.propagate_reject = function (d) {
    "use strict";
    var id = d.id;
    d3.selectAll(".D" + id + ":not(.rej)")
        .classed({"rej": true, "prej": false, "pacc": false})
        .each(GambleHex.propagate_reject);
    d3.selectAll(".d" + id + ":not(.rej):not(.accd):not(.acc):not(.prej)")
        .classed("prej", true);
};
GambleHex.propagate = function (d) {
    "use strict";
    var id = d.id;
    d3.selectAll(".N" + id + ":not(.rej)")
        .classed({"rej": true, "prej": false, "pacc": false})
        .each(GambleHex.propagate_reject);
    d3.selectAll(".d" + id + ":not(.acc)")
        .classed({"acc": true, "prej": false, "pacc": false})
        .each(GambleHex.propagate);
    d3.selectAll(".D" + id + ":not(.acc):not(.accd):not(.rej):not(.pacc)")
        .classed("pacc", true);
};
GambleHex.paccinforder = function (a, b) {
    "use strict";
    // gambles with low beta and then lambda values in front
    return a.beta !== b.beta ? a.beta - b.beta : a.lambda - b.lambda;
};
GambleHex.paccinf = function (n, c, A) {
    "use strict";
    var selection = d3.selectAll(".pacc");
    if (selection.empty()) { return; }
    selection.sort(GambleHex.paccinforder)
        .each(function (d) {
            var node = d3.select(this),
                b,
                lpr;
            if (node.classed("acc")) { return; }
            b = [d.W, d.D, d.L].concat(numeric.rep([n], 0));
            lpr = numeric.solveLP(c, A, b).solution[0];
            if (numeric.trunc(lpr, 1e-6) >= 0) {
                node.classed("acc", true).each(GambleHex.propagate);
            }
            node.classed("pacc", false);
        });
    GambleHex.paccinf(n, c, A);
};
GambleHex.prejinforder = function (a, b) {
    "use strict";
    // gambles with high beta and then lambda values in front
    return -GambleHex.paccinforder(a, b);
};
GambleHex.prejinf = function (n, c, A) {
    "use strict";
    var selection = d3.selectAll(".prej");
    if (selection.empty()) { return; }
    selection.sort(GambleHex.prejinforder)
        .each(function (d) {
            var node = d3.select(this),
                b,
                upr;
            if (node.classed("rej")) { return; }
            b = [-d.W, -d.D, -d.L].concat(numeric.rep([n], 0));
            upr = -numeric.solveLP(c, A, b).solution[0];
            if (numeric.trunc(upr, 1e-6) < 0) {
                node.classed("rej", true).each(GambleHex.propagate_reject);
            }
            node.classed("prej", false);
        });
    GambleHex.prejinf(n, c, A);
};
GambleHex.infer = function (selection) {
    "use strict";
    selection.each(GambleHex.propagate);
    /* start linear programming */
    var n = selection.size(),
        i = 0,
        D,
        A,
        c;
    if (n < 1) { return; }
    D = numeric.rep([3, n], 0);
    selection.each(function (d) {
        numeric.setBlock(D, [0, i], [2, i], [[d.W], [d.D], [d.L]]);
        i += 1;
    });
    A = numeric.rep([3 + n, 1 + n], 0);
    numeric.setBlock(A, [0, 0], [2, 0], numeric.rep([3, 1], 1));
    numeric.setBlock(A, [3, 1], [3 + n - 1, n],
                    numeric.neg(numeric.identity(n)));
    numeric.setBlock(A, [0, 1], [2, n], D);
    c = numeric.rep([1 + n], 0);
    c[0] = -1;
    GambleHex.paccinf(n, c, A);
    GambleHex.prejinf(n, c, A);
};

/* Import and export functions */
GambleHex.accdImport = function (accdobjs) {  // [{"W": w, "D": d, "L": l}, ...]
    "use strict";
    accdobjs.forEach(function (accdobj) {
        var dec2frac = GambleHex.dec2frac;
        d3.select("#".concat(
            dec2frac(accdobj.W),
            dec2frac(accdobj.D),
            dec2frac(accdobj.L)
        )).classed("accd", true);
    });
    d3.selectAll(".accd").call(GambleHex.infer);
};
GambleHex.accdExport = function () {
    "use strict";
    var accdobjs = [];
    d3.selectAll(".accd").each(function (d) {
        accdobjs.push({"W": d.W, "D": d.D, "L": d.L});
    });
    return accdobjs;  // [{"W": w, "D": d, "L": l}, ...]
};
GambleHex.assign = function (w, d, l) {
    "use strict";
    var crds = GambleHex.coords(w, d, l);
    d3.select("#assigned")
        .attr("cx", crds.x)
        .attr("cy", crds.y)
        .attr("r", 30 / (2 * GambleHex.lg.max + 2))
        .each(function (datum) {
            datum.W = w;
            datum.D = d;
            datum.L = l;
        });
    console.log(Date.now(), "assign", [w, d, l]);
};


/* *** Draw and initialize the interface *** */
window.onload = function () {
    "use strict";
    var svg,
        values,
        sure_gain,
        sure_loss,
        nodes,
        assigned;

    (function () {  // Static aspects
        var offset = GambleHex.offset,
            coords = GambleHex.coords,
            sure = d3.map({  // Array of extreme sure gambles' coordinates
                "W": coords(0, -1, -1),
                "NL": coords(0, 0, -1),
                "D": coords(-1, 0, -1),
                "NW": coords(-1, 0, 0),
                "L": coords(-1, -1, 0),
                "ND": coords(0, -1, 0)
            }),
            unsure = d3.map({  // Array of extreme unsure gambles' coordinates
                "W": coords(1, 0, 0),
                "NL": coords(1, 1, 0),
                "D": coords(0, 1, 0),
                "NW": coords(0, 1, 1),
                "L": coords(0, 0, 1),
                "ND": coords(1, 0, 1)
            }),
            lineFunction = d3.svg.line()  // line data accessor
                                    .x(function (d) { return d.x; })
                                    .y(function (d) { return d.y; })
                                    .interpolate("linear");
        svg = d3.select("#playfield").append("svg")  // Create the SVG
                .attr("width", (offset / 3) + "ex")
                .attr("height", (offset / 3) + "ex")
                .attr("viewBox", "0 0 " + 2 * offset + ' ' + 2 * offset);

        (function () {  // Draw all axes
            /* Draw radial axes */
            unsure.values().forEach(function (coords) {
                svg.append("path")
                    .attr("d", lineFunction([GambleHex.origin, coords]))
                    .classed("axis", true);
            });

            /* Draw conditional axes */
            d3.map({
                "NW": [coords(-1, 0, +Infinity), coords(-1, +Infinity, 0)],
                "ND": [coords(0, -1, +Infinity), coords(+Infinity, -1, 0)],
                "NL": [coords(0, +Infinity, -1), coords(+Infinity, 0, -1)]
            }).forEach(function (event, coordsarr) {
                coordsarr.forEach(function (coords) {
                    svg.append("path")
                            .attr("d", lineFunction([sure.get(event), coords]))
                            .classed("axis", true);
                });
            });

            /* Draw even odds line */
            svg.append("path")
                .attr("d", lineFunction([
                    coords(1, -1, -1),
                    coords(1, 1, -1),
                    coords(-1, 1, -1),
                    coords(-1, 1, 1),
                    coords(-1, -1, 1),
                    coords(1, -1, 1)
                ]) + 'Z')
                .classed("axis", true);
        }());

        (function () {  // Initialize the labels and label values
            var lbl = GambleHex.lbl,
                labels = d3.map(),
                label_values;
            function labelscale(val) { return 1.3 * (val - offset) + offset; }
            unsure.forEach(function (event, coords) {
                labels.set(event, {
                    'x': labelscale(coords.x),
                    'y': labelscale(coords.y)
                });
            });
            /* Draw the initial labels */
            svg.selectAll(".label")
                .data(labels.keys()).enter()
                .append("text")
                .attr("id", function (event) { return event; })
                .attr("x", function (event) { return labels.get(event).x; })
                .attr("y", function (event) { return labels.get(event).y; })
                .text(function (event) { return lbl[event]; })
                .classed("label", true);
            svg.selectAll("#W, #D, #L").classed("elementary", true);
            svg.selectAll("#NW, #ND, #NL").classed("nonelementary", true);
            /* Create the initial gamble value containers */
            label_values = d3.map();
            ["W", "D", "L"].forEach(function (event) {
                var crds = labels.get(event);
                label_values.set(event, {'x': crds.x, 'y': crds.y + 22 });
            });
            values = svg.selectAll(".value")
                .data(label_values.keys()).enter()
                .append("text")
                .attr("id", function (state) { return state; })
                .attr("x", function (state) {
                    return label_values.get(state).x;
                })
                .attr("y", function (state) {
                    return label_values.get(state).y;
                })
                .text("±0")
                .classed("value", true);
        }());

        (function () {  // Draw hexagons
            /* Draw outer hexagon */
            sure_gain = svg.append("path")
                            .attr("d", lineFunction(unsure.values()) + 'Z')
                            .classed("sure_gain", true);

            /* Draw inner hexagon */
            sure_loss = svg.append("path")
                            .attr("d", lineFunction(sure.values()) + 'Z')
                            .classed("sure_loss", true);
        }());

        (function () {  // Draw grid
            var lg = GambleHex.lg,
                griddata = GambleHex.griddata,
                tint = d3.interpolateRgb(
                    // colorscale for unresolved gambles
                    d3.rgb("tomato"),
                    d3.rgb("cornflowerblue")
                );
            nodes = svg.selectAll(".gamble")
                .data(GambleHex.grid.values()).enter()
                .append("circle")
                .attr("id", function (d) { return d.id; })
                .attr("cx", function (d) {
                    return griddata.get(d.id).x;
                })
                .attr("cy", function (d) {
                    return griddata.get(d.id).y;
                })
                .attr("r", 41 / (2 * lg.max + 2))
                .classed("gamble", true)
                .each(function (d) {
                    var node = d3.select(this),
                        datum = griddata.get(d.id);
                    datum.dominating.forEach(function (id) {
                        node.classed('D' + id, true);
                    });
                    datum.dominated.forEach(function (id) {
                        node.classed('d' + id, true);
                    });
                    datum.negdominating.forEach(function (id) {
                        node.classed('N' + id, true);
                    });
                })
                .style("fill", function (d) { return tint(d.beta); });
        }());

        (function () {  // Assigned gamble (initialized unassigned)
            assigned = svg.selectAll("#assigned")
                .data([{"W": 0, "D": 0, "L": 0}]).enter()
                .append("circle")
                .attr("id", "assigned");
        }());
    }());

    (function () {  // Dynamic aspects
        var dec2frac = GambleHex.dec2frac;
        /* Function to set class that indicates sign of a gamble value */
        function valuesign(valstr) {
            var sign = valstr.charAt(0);
            return {'positive': sign === '+', 'negative': sign === '−'};
        }

        (function () {  // Hexagons
            function statusquovals() {
                values.text("±0").classed(valuesign("0"));
            }
            sure_gain.on("mouseover", statusquovals);
            function surelossvals() {
                values.text("−1").classed(valuesign("−1"));
            }
            sure_loss.on("mouseover", surelossvals);
        }());

        (function () {  // Grid
            function dec2output(val) {
                return dec2frac(val).replace('＋', '+').replace('÷', '/');
                    // we replace some unicode symbols by nicer ones when
                    // outputting
            }
            function generalvals(d) {
                values.each(function (state) {
                    var valstr = dec2output(d[state]);
                    d3.select(this)
                        .text(valstr)
                        .classed(valuesign(valstr));
                });
            }
            function toggle_accept() {
                var node = d3.select(this),
                    accy = !node.classed("accd");
                if (!node.filter(".acc:not(.accd)").empty()) {
                    console.log(Date.now(), node.attr("id"), "acc");
                    return;
                }
                if (!node.filter(".rej:not(.accd)").empty()) {
                    console.log(Date.now(), node.attr("id"), "rej");
                    return;
                }
                node.classed("accd", accy);
                console.log(Date.now(), node.attr("id"), "accd", accy);
                d3.selectAll(".acc, .rej, .pacc, .prej")
                    .classed({
                        "acc": false,
                        "rej": false,
                        "pacc": false,
                        "prej": false
                    });
                d3.selectAll(".accd").call(GambleHex.infer);
            }
            nodes.on("mouseover", generalvals).on("click", toggle_accept);
        }());

        (function () {  // Assigned gamble
            function assignedvals(d) {
                values.classed("assigned", true).each(function (state) {
                    var valstr = numeric.trunc(d[state], 1e-2);
                    d3.select(this).text(valstr);
                });
            }
            assigned.on("mouseover", assignedvals).on("mouseout", function () {
                values.classed("assigned", false);
            });
        }());
    }());
};
